@@ -60,13 +60,11 @@ var wopiSrc = getParameterByName('WOPISrc');
 if (wopiSrc !== '' && access_token !== '') {
 	var wopiParams = { 'access_token': access_token, 'access_token_ttl': access_token_ttl };
 }
-
-var filePath = getParameterByName('file_path');
-var title = getParameterByName('title');
-if (title === '') {
-    title = decodeURIComponent(filePath.substring(filePath.lastIndexOf('/')+1));
+else if (wopiSrc !== '' && access_header !== '') {
+	var wopiParams = { 'access_header': access_header };
 }
 
+var filePath = getParameterByName('file_path');
 var permission = getParameterByName('permission') || 'edit';
 var timestamp = getParameterByName('timestamp');
 // Shows close button if non-zero value provided
@@ -88,7 +86,6 @@ if (host === '') {
 // TODO: Get rid of these globals
 global.closebutton = closebutton;
 global.revHistoryEnabled = revHistoryEnabled;
-global.title = title;
 global.errorMessages = errorMessages;
 var docURL, docParams;
 var isWopi = false;
@@ -97,11 +94,10 @@ if (wopiSrc != '') {
 	docParams = wopiParams;
 	isWopi = true;
 } else {
-    docURL = filePath;
+	docURL = filePath;
 	docParams = {};
 }
 
-document.title = title;
 var map = L.map('map', {
 	server: host,
 	doc: docURL,
@@ -111,6 +107,7 @@ var map = L.map('map', {
 	documentContainer: 'document-container',
 	debug: debugMode,
 	wopi: isWopi,
+	wopiSrc: wopiSrc,
 	alwaysActive: alwaysActive,
 	idleTimeoutSecs: idleTimeoutSecs,  // Dim when user is idle.
 	outOfFocusTimeoutSecs: outOfFocusTimeoutSecs // Dim after switching tabs.
@@ -128,6 +125,7 @@ map.addControl(L.control.columnHeader());
 map.addControl(L.control.rowHeader());
 map.addControl(L.control.contextMenu());
 map.addControl(L.control.menubar());
+map.loadDocument();
 
 window.addEventListener('beforeunload', function () {
 	if (global.map && global.map._socket) {
