@@ -63,6 +63,7 @@ L.AnnotationManager = L.Class.extend({
 				weight: 2,
 				opacity: 0.25
 			});
+			L.DomEvent.on(comment.textSelected, 'contextmenu', L.DomEvent.preventDefault);
 			comment.textSelected.on('click', function(e) {
 				// Simulate a click at this position in the document
 				var latlng = this._map.mouseEventToLatLng(e.originalEvent);
@@ -100,6 +101,7 @@ L.AnnotationManager = L.Class.extend({
 				fillOpacity: 0,
 				opacity: 0
 			});
+			L.DomEvent.on(redline.textSelected, 'contextmenu', L.DomEvent.preventDefault);
 			redline.textSelected.on('click', function(e) {
 				// Simulate a click at this position in the document
 				var latlng = this._map.mouseEventToLatLng(e.originalEvent);
@@ -123,6 +125,9 @@ L.AnnotationManager = L.Class.extend({
 		for (var index in comments) {
 			comment = comments[index];
 			this.adjustComment(comment);
+			if (comment.author in this._map._viewInfoByUserName) {
+				comment.avatar = this._map._viewInfoByUserName[comment.author].userextrainfo.avatar;
+			}
 			this._items.push(L.annotation(this._map.options.maxBounds.getSouthEast(), comment).addTo(this._map));
 		}
 		if (this._items.length > 0) {
@@ -148,6 +153,9 @@ L.AnnotationManager = L.Class.extend({
 				// something wrong in this redline, skip this one
 				continue;
 			}
+			if (changecomment.author in this._map._viewInfoByUserName) {
+				changecomment.avatar = this._map._viewInfoByUserName[changecomment.author].userextrainfo.avatar;
+			}
 			this._items.push(L.annotation(this._map.options.maxBounds.getSouthEast(), changecomment).addTo(this._map));
 		}
 		if (this._items.length > 0) {
@@ -519,6 +527,14 @@ L.AnnotationManager = L.Class.extend({
 		var id;
 		var changetrack = obj.redline ? true : false;
 		var action = changetrack ? obj.redline.action : obj.comment.action;
+
+		if (changetrack && obj.redline.author in this._map._viewInfoByUserName) {
+			obj.redline.avatar = this._map._viewInfoByUserName[obj.redline.author].userextrainfo.avatar;
+		}
+		else if (!changetrack && obj.comment.author in this._map._viewInfoByUserName) {
+			obj.comment.avatar = this._map._viewInfoByUserName[obj.comment.author].userextrainfo.avatar;
+		}
+
 		if (action === 'Add') {
 			if (changetrack) {
 				if (!this.adjustRedLine(obj.redline)) {
