@@ -351,7 +351,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 
 		var tilePositionsX = '';
 		var tilePositionsY = '';
-		var oldHashes = '';
+		var oldWireIds = '';
 		var needsNewTiles = false;
 
 		for (var key in this._tiles) {
@@ -375,14 +375,14 @@ L.ImpressTileLayer = L.TileLayer.extend({
 						tilePositionsY += ',';
 					}
 					tilePositionsY += tileTopLeft.y;
-					if (oldHashes !== '') {
-						oldHashes += ',';
+					if (oldWireIds !== '') {
+						oldWireIds += ',';
 					}
-					if (this._tiles[key].oldhash === undefined) {
-						oldHashes += '0';
+					if (this._tiles[key].oldWireId === undefined) {
+						oldWireIds += '0';
 					}
 					else {
-						oldHashes += this._tiles[key].oldhash;
+						oldWireIds += this._tiles[key].oldWireId;
 					}
 					needsNewTiles = true;
 					if (this._debug) {
@@ -407,7 +407,7 @@ L.ImpressTileLayer = L.TileLayer.extend({
 				'tileposy=' + tilePositionsY + ' ' +
 				'tilewidth=' + this._tileWidthTwips + ' ' +
 				'tileheight=' + this._tileHeightTwips + ' ' +
-				'oldhash=' + oldHashes;
+				'oldwid=' + oldWireIds;
 
 			this._map._socket.sendMessage(message, '');
 			if (this._debug) {
