@@ -26,9 +26,9 @@
 #include "Unit.hpp"
 #include "Util.hpp"
 
-void Document::addView(const std::string& sessionId, const std::string& userName)
+void Document::addView(const std::string& sessionId, const std::string& userName, const std::string& userId)
 {
-    const auto ret = _views.emplace(sessionId, View(sessionId, userName));
+    const auto ret = _views.emplace(sessionId, View(sessionId, userName, userId));
     if (!ret.second)
     {
         LOG_WRN("View with SessionID [" << sessionId << "] already exists.");
@@ -50,7 +50,7 @@ int Document::expireView(const std::string& sessionId)
         if (--_activeViews == 0)
             _end = std::time(nullptr);
     }
-    this->takeSnapshot();
+    takeSnapshot();
 
     return _activeViews;
 }
@@ -61,12 +61,12 @@ std::pair<std::time_t, std::string> Document::getSnapshot() const
     std::ostringstream oss;
     oss << "{";
     oss << "\"creationTime\"" << ":" << ct << ",";
-    oss << "\"memoryDirty\"" << ":" << this->getMemoryDirty() << ",";
-    oss << "\"activeViews\"" << ":" << this->getActiveViews() << ",";
+    oss << "\"memoryDirty\"" << ":" << getMemoryDirty() << ",";
+    oss << "\"activeViews\"" << ":" << getActiveViews() << ",";
 
     oss << "\"views\"" << ":[";
     std::string separator;
-    for (auto view : this->getViews())
+    for (auto view : getViews())
     {
         oss << separator << "\"";
         if(view.second.isExpired())
@@ -78,7 +78,7 @@ std::pair<std::time_t, std::string> Document::getSnapshot() const
     }
     oss << "],";
 
-    oss << "\"lastActivity\"" << ":" << this->_lastActivity;
+    oss << "\"lastActivity\"" << ":" << _lastActivity;
     oss << "}";
     return std::make_pair(ct, oss.str());
 }
@@ -87,11 +87,11 @@ const std::string Document::getHistory() const
 {
     std::ostringstream oss;
     oss << "{";
-    oss << "\"docKey\"" << ":\"" << this->_docKey << "\",";
-    oss << "\"filename\"" << ":\"" << this->getFilename() << "\",";
-    oss << "\"start\"" << ":" << this->_start << ",";
-    oss << "\"end\"" << ":" << this->_end << ",";
-    oss << "\"pid\"" << ":" << this->getPid() << ",";
+    oss << "\"docKey\"" << ":\"" << _docKey << "\",";
+    oss << "\"filename\"" << ":\"" << getFilename() << "\",";
+    oss << "\"start\"" << ":" << _start << ",";
+    oss << "\"end\"" << ":" << _end << ",";
+    oss << "\"pid\"" << ":" << getPid() << ",";
     oss << "\"snapshots\"" << ":[";
     std::string separator;
     for (auto s : _snapshots)
@@ -105,7 +105,7 @@ const std::string Document::getHistory() const
 
 void Document::takeSnapshot()
 {
-    auto p = this->getSnapshot();
+    auto p = getSnapshot();
     auto insPoint = _snapshots.upper_bound(p.first);
     _snapshots.insert(insPoint, p);
 }
@@ -114,13 +114,13 @@ std::string Document::to_string() const
 {
     std::ostringstream oss;
     std::string encodedFilename;
-    Poco::URI::encode(this->getFilename(), " ", encodedFilename);
-    oss << this->getPid() << ' '
+    Poco::URI::encode(getFilename(), " ", encodedFilename);
+    oss << getPid() << ' '
         << encodedFilename << ' '
-        << this->getActiveViews() << ' '
-        << this->getMemoryDirty() << ' '
-        << this->getElapsedTime() << ' '
-        << this->getIdleTime() << ' ';
+        << getActiveViews() << ' '
+        << getMemoryDirty() << ' '
+        << getElapsedTime() << ' '
+        << getIdleTime() << ' ';
     return oss.str();
 }
 
@@ -178,8 +178,8 @@ void AdminModel::assertCorrectThread() const
 
 AdminModel::~AdminModel()
 {
-    Log::debug("History:\n\n" + getAllHistory() + "\n");
-    Log::info("AdminModel dtor.");
+    LOG_DBG("History:\n\n" << getAllHistory() << '\n');
+    LOG_INF("AdminModel dtor.");
 }
 
 std::string AdminModel::getAllHistory() const
@@ -187,7 +187,7 @@ std::string AdminModel::getAllHistory() const
     std::ostringstream oss;
     oss << "{ \"documents\" : [";
     std::string separator1;
-    for (auto d : _documents)
+    for (const auto& d : _documents)
     {
         oss << separator1;
         oss << d.second.getHistory();
@@ -195,7 +195,7 @@ std::string AdminModel::getAllHistory() const
     }
     oss << "], \"expiredDocuments\" : [";
     separator1 = "";
-    for (auto ed : _expiredDocuments)
+    for (const auto& ed : _expiredDocuments)
     {
         oss << separator1;
         oss << ed.second.getHistory();
@@ -238,6 +238,18 @@ std::string AdminModel::query(const std::string& command)
     {
         return std::to_string(_cpuStatsSize);
     }
+    else if (token == "sent_activity")
+    {
+        return getSentActivity();
+    }
+    else if (token == "recv_activity")
+    {
+        return getRecvActivity();
+    }
+    else if (token == "net_stats_size")
+    {
+        return std::to_string(std::max(_sentStatsSize, _recvStatsSize));
+    }
 
     return std::string("");
 }
@@ -271,11 +283,36 @@ unsigned AdminModel::getKitsMemoryUsage()
     return totalMem;
 }
 
+size_t AdminModel::getKitsJiffies()
+{
+    assertCorrectThread();
+
+    size_t totalJ = 0;
+    for (auto& it : _documents)
+    {
+        if (!it.second.isExpired())
+        {
+            const auto pid = it.second.getPid();
+            if (pid > 0)
+            {
+                unsigned newJ = Util::getCpuUsage(pid);
+                unsigned prevJ = it.second.getLastJiffies();
+                if(newJ >= prevJ)
+                {
+                    totalJ += (newJ - prevJ);
+                    it.second.setLastJiffies(newJ);
+                }
+            }
+        }
+    }
+    return totalJ;
+}
+
 void AdminModel::subscribe(int sessionId, const std::weak_ptr<WebSocketHandler>& ws)
 {
     assertCorrectThread();
 
-    const auto ret = _subscribers.emplace(sessionId, Subscriber(sessionId, ws));
+    const auto ret = _subscribers.emplace(sessionId, Subscriber(ws));
     if (!ret.second)
     {
         LOG_WRN("Subscriber already exists");
@@ -324,6 +361,28 @@ void AdminModel::addCpuStats(unsigned cpuUsage)
     notify("cpu_stats " + std::to_string(cpuUsage));
 }
 
+void AdminModel::addSentStats(uint64_t sent)
+{
+    assertCorrectThread();
+
+    _sentStats.push_back(sent);
+    if (_sentStats.size() > _sentStatsSize)
+        _sentStats.pop_front();
+
+    notify("sent_activity " + std::to_string(sent));
+}
+
+void AdminModel::addRecvStats(uint64_t recv)
+{
+    assertCorrectThread();
+
+    _recvStats.push_back(recv);
+    if (_recvStats.size() > _recvStatsSize)
+        _recvStats.pop_front();
+
+    notify("recv_activity " + std::to_string(recv));
+}
+
 void AdminModel::setCpuStatsSize(unsigned size)
 {
     assertCorrectThread();
@@ -373,19 +432,51 @@ void AdminModel::notify(const std::string& message)
     }
 }
 
+void AdminModel::addBytes(const std::string& docKey, uint64_t sent, uint64_t recv)
+{
+    assertCorrectThread();
+
+    auto doc = _documents.find(docKey);
+    if(doc != _documents.end())
+        doc->second.addBytes(sent, recv);
+
+    _sentBytesTotal += sent;
+    _recvBytesTotal += recv;
+}
+
+void AdminModel::modificationAlert(const std::string& docKey, Poco::Process::PID pid, bool value)
+{
+    assertCorrectThread();
+
+    auto doc = _documents.find(docKey);
+    if(doc != _documents.end())
+    {
+        doc->second.setModified(value);
+    }
+
+    std::ostringstream oss;
+    oss << "modifications "
+        << pid << ' '
+        << (value?"Yes":"No");
+
+    notify(oss.str());
+}
+
 void AdminModel::addDocument(const std::string& docKey, Poco::Process::PID pid,
                              const std::string& filename, const std::string& sessionId,
-                             const std::string& userName)
+                             const std::string& userName, const std::string& userId)
 {
     assertCorrectThread();
 
     const auto ret = _documents.emplace(docKey, Document(docKey, pid, filename));
     ret.first->second.takeSnapshot();
-    ret.first->second.addView(sessionId, userName);
+    ret.first->second.addView(sessionId, userName, userId);
     LOG_DBG("Added admin document [" << docKey << "].");
 
     std::string encodedUsername;
     std::string encodedFilename;
+    std::string encodedUserId;
+    Poco::URI::encode(userId, " ", encodedUserId);
     Poco::URI::encode(filename, " ", encodedFilename);
     Poco::URI::encode(userName, " ", encodedUsername);
 
@@ -395,7 +486,8 @@ void AdminModel::addDocument(const std::string& docKey, Poco::Process::PID pid,
         << pid << ' '
         << encodedFilename << ' '
         << sessionId << ' '
-        << encodedUsername << ' ';
+        << encodedUsername << ' '
+        << encodedUserId << ' ';
 
     // We have to wait until the kit sends us its PSS.
     // Here we guestimate until we get an update.
@@ -495,6 +587,32 @@ std::string AdminModel::getCpuStats()
     return oss.str();
 }
 
+std::string AdminModel::getSentActivity()
+{
+    assertCorrectThread();
+
+    std::ostringstream oss;
+    for (const auto& i: _sentStats)
+    {
+        oss << i << ',';
+    }
+
+    return oss.str();
+}
+
+std::string AdminModel::getRecvActivity()
+{
+    assertCorrectThread();
+
+    std::ostringstream oss;
+    for (const auto& i: _recvStats)
+    {
+        oss << i << ',';
+    }
+
+    return oss.str();
+}
+
 unsigned AdminModel::getTotalActiveViews()
 {
     assertCorrectThread();
@@ -511,6 +629,25 @@ unsigned AdminModel::getTotalActiveViews()
     return numTotalViews;
 }
 
+std::list<DocBasicInfo> AdminModel::getDocumentsSortedByIdle() const
+{
+    std::list<DocBasicInfo> docList;
+    for (const auto& it: _documents)
+    {
+        docList.emplace_back(it.second.getDocKey(),
+                             it.second.getIdleTime(),
+                             !it.second.getModifiedStatus(),
+                             it.second.getMemoryDirty());
+    }
+
+    // Sort the list by idle times;
+    docList.sort([](const DocBasicInfo& a, const DocBasicInfo& b) {
+            return a._idleTime > b._idleTime;
+        });
+
+    return docList;
+}
+
 std::string AdminModel::getDocuments() const
 {
     assertCorrectThread();
@@ -532,6 +669,7 @@ std::string AdminModel::getDocuments() const
                 << "\"memory\"" << ':' << it.second.getMemoryDirty() << ','
                 << "\"elapsedTime\"" << ':' << it.second.getElapsedTime() << ','
                 << "\"idleTime\"" << ':' << it.second.getIdleTime() << ','
+                << "\"modified\"" << ':' << '"' << (it.second.getModifiedStatus() ? "Yes" : "No") << '"' << ','
                 << "\"views\"" << ':' << '[';
             viewers = it.second.getViews();
             std::string separator;
@@ -540,6 +678,7 @@ std::string AdminModel::getDocuments() const
                 if(!viewIt.second.isExpired()) {
                     oss << separator << '{'
                         << "\"userName\"" << ':' << '"' << viewIt.second.getUserName() << '"' << ','
+                        << "\"userId\"" << ':' << '"' << viewIt.second.getUserId() << '"' << ','
                         << "\"sessionid\"" << ':' << '"' << viewIt.second.getSessionId() << '"' << '}';
                         separator = ',';
                 }
