@@ -11,30 +11,32 @@
 
 #include "DocumentBroker.hpp"
 
+#include <atomic>
 #include <cassert>
 #include <chrono>
 #include <ctime>
 #include <fstream>
 #include <sstream>
 
+#include <Poco/DigestStream.h>
+#include <Poco/Exception.h>
 #include <Poco/JSON/Object.h>
 #include <Poco/Path.h>
 #include <Poco/SHA1Engine.h>
-#include <Poco/DigestStream.h>
 #include <Poco/StreamCopier.h>
 #include <Poco/StringTokenizer.h>
 
 #include "Admin.hpp"
 #include "ClientSession.hpp"
 #include "Exceptions.hpp"
-#include "Message.hpp"
-#include "Protocol.hpp"
 #include "LOOLWSD.hpp"
-#include "Log.hpp"
+#include "SenderQueue.hpp"
 #include "Storage.hpp"
 #include "TileCache.hpp"
-#include "SenderQueue.hpp"
-#include "Unit.hpp"
+#include "common/Log.hpp"
+#include "common/Message.hpp"
+#include "common/Protocol.hpp"
+#include "common/Unit.hpp"
 
 using namespace LOOLProtocol;
 
@@ -146,6 +148,7 @@ DocumentBroker::DocumentBroker(const std::string& uri,
     _docId(Util::encodeId(DocBrokerId++, 3)),
     _childRoot(childRoot),
     _cacheRoot(getCachePath(uriPublic.toString())),
+    _documentChangedInStorage(false),
     _lastSaveTime(std::chrono::steady_clock::now()),
     _lastSaveRequestTime(std::chrono::steady_clock::now() - std::chrono::milliseconds(COMMAND_TIMEOUT_MS)),
     _markToDestroy(false),
@@ -158,6 +161,7 @@ DocumentBroker::DocumentBroker(const std::string& uri,
     _cursorHeight(0),
     _poll(new DocumentBrokerPoll("docbroker_" + _docId, *this)),
     _stop(false),
+    _closeReason("stopped"),
     _tileVersion(0),
     _debugRenderedTileCount(0)
 {
@@ -227,7 +231,10 @@ void DocumentBroker::pollThread()
     static const bool AutoSaveEnabled = !std::getenv("LOOL_NO_AUTOSAVE");
     static const size_t IdleDocTimeoutSecs = LOOLWSD::getConfigValue<int>(
                                                       "per_document.idle_timeout_secs", 3600);
-    std::string closeReason = "stopped";
+    // Used to accumulate B/W deltas.
+    uint64_t adminSent = 0;
+    uint64_t adminRecv = 0;
+    auto lastBWUpdateTime = std::chrono::steady_clock::now();
 
     // Main polling loop goodness.
     while (!_stop && _poll->continuePolling() && !TerminationFlag)
@@ -235,6 +242,23 @@ void DocumentBroker::pollThread()
         _poll->poll(SocketPoll::DefaultPollTimeoutMs);
 
         const auto now = std::chrono::steady_clock::now();
+
+        if (std::chrono::duration_cast<std::chrono::milliseconds>
+                    (now - lastBWUpdateTime).count() >= 5 * 1000)
+        {
+            lastBWUpdateTime = now;
+            uint64_t sent, recv;
+            getIOStats(sent, recv);
+            // send change since last notification.
+            Admin::instance().addBytes(getDocKey(),
+                                       // connection drop transiently reduces this.
+                                       (sent > adminSent ? (sent - adminSent): uint64_t(0)),
+                                       (recv > adminRecv ? (recv - adminRecv): uint64_t(0)));
+            LOG_INF("Doc [" << _docKey << "] added sent: " << sent << " recv: " << recv << " bytes to totals");
+            adminSent = sent;
+            adminRecv = recv;
+        }
+
         if (_lastSaveTime < _lastSaveRequestTime &&
             std::chrono::duration_cast<std::chrono::milliseconds>
                     (now - _lastSaveRequestTime).count() <= COMMAND_TIMEOUT_MS)
@@ -245,7 +269,7 @@ void DocumentBroker::pollThread()
 
         if (ShutdownRequestFlag)
         {
-            closeReason = "recycling";
+            _closeReason = "recycling";
             _stop = true;
         }
         else if (AutoSaveEnabled && !_stop &&
@@ -264,7 +288,7 @@ void DocumentBroker::pollThread()
         {
             LOG_INF("Terminating " << (idle ? "idle" : "dead") <<
                     " DocumentBroker for docKey [" << getDocKey() << "].");
-            closeReason = (idle ? "idle" : "dead");
+            _closeReason = (idle ? "idle" : "dead");
             _stop = true;
         }
     }
@@ -287,7 +311,7 @@ void DocumentBroker::pollThread()
     }
 
     // Terminate properly while we can.
-    terminateChild(closeReason, false);
+    terminateChild(_closeReason);
 
     // Stop to mark it done and cleanup.
     _poll->stop();
@@ -392,6 +416,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
             LOG_ERR("Failed to create Storage instance for [" << _docKey << "] in " << jailPath.toString());
             return false;
         }
+
         firstInstance = true;
     }
 
@@ -399,15 +424,20 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
 
     // Call the storage specific fileinfo functions
     std::string userid, username;
+    std::string userExtraInfo;
+    std::string watermarkText;
     std::chrono::duration<double> getInfoCallDuration(0);
     WopiStorage* wopiStorage = dynamic_cast<WopiStorage*>(_storage.get());
     if (wopiStorage != nullptr)
     {
-        std::unique_ptr<WopiStorage::WOPIFileInfo> wopifileinfo = wopiStorage->getWOPIFileInfo(session->getAccessToken());
+        std::unique_ptr<WopiStorage::WOPIFileInfo> wopifileinfo = wopiStorage->getWOPIFileInfo(session->getAuthorization());
         userid = wopifileinfo->_userid;
         username = wopifileinfo->_username;
+        userExtraInfo = wopifileinfo->_userExtraInfo;
+        watermarkText = wopifileinfo->_watermarkText;
 
-        if (!wopifileinfo->_userCanWrite)
+        if (!wopifileinfo->_userCanWrite ||
+            LOOLWSD::IsViewFileExtension(wopiStorage->getFileExtension()))
         {
             LOG_DBG("Setting the session as readonly");
             session->setReadOnly();
@@ -434,12 +464,15 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         if (wopifileinfo->_disableExport)
             wopifileinfo->_hideExportOption = true;
 
+        wopiInfo->set("BaseFileName", wopiStorage->getFileInfo()._filename);
         wopiInfo->set("HidePrintOption", wopifileinfo->_hidePrintOption);
         wopiInfo->set("HideSaveOption", wopifileinfo->_hideSaveOption);
         wopiInfo->set("HideExportOption", wopifileinfo->_hideExportOption);
         wopiInfo->set("DisablePrint", wopifileinfo->_disablePrint);
         wopiInfo->set("DisableExport", wopifileinfo->_disableExport);
         wopiInfo->set("DisableCopy", wopifileinfo->_disableCopy);
+        wopiInfo->set("DisableInactiveMessages", wopifileinfo->_disableInactiveMessages);
+        wopiInfo->set("UserCanNotWriteRelative", wopifileinfo->_userCanNotWriteRelative);
 
         std::ostringstream ossWopiInfo;
         wopiInfo->stringify(ossWopiInfo);
@@ -451,7 +484,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         // Mark the session as 'Document owner' if WOPI hosts supports it
         if (userid == _storage->getFileInfo()._ownerId)
         {
-            LOG_DBG("Session [" + sessionId + "] is the document owner");
+            LOG_DBG("Session [" << sessionId << "] is the document owner");
             session->setDocumentOwner(true);
         }
 
@@ -468,12 +501,25 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
             std::unique_ptr<LocalStorage::LocalFileInfo> localfileinfo = localStorage->getLocalFileInfo();
             userid = localfileinfo->_userid;
             username = localfileinfo->_username;
+
+            if (LOOLWSD::IsViewFileExtension(localStorage->getFileExtension()))
+            {
+                LOG_DBG("Setting the session as readonly");
+                session->setReadOnly();
+            }
         }
     }
 
+#if ENABLE_SUPPORT_KEY
+    if (!LOOLWSD::OverrideWatermark.empty())
+        watermarkText = LOOLWSD::OverrideWatermark;
+#endif
+
     LOG_DBG("Setting username [" << username << "] and userId [" << userid << "] for session [" << sessionId << "]");
     session->setUserId(userid);
     session->setUserName(username);
+    session->setUserExtraInfo(userExtraInfo);
+    session->setWatermarkText(watermarkText);
 
     // Basic file information was stored by the above getWOPIFileInfo() or getLocalFileInfo() calls
     const auto fileInfo = _storage->getFileInfo();
@@ -486,29 +532,86 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     if (firstInstance)
     {
         _documentLastModifiedTime = fileInfo._modifiedTime;
-        LOG_DBG("Document timestamp: " << Poco::DateTimeFormatter::format(Poco::DateTime(_documentLastModifiedTime),
-                                                                          Poco::DateTimeFormat::ISO8601_FORMAT));
+        Log::debug() << "Document timestamp: " << _documentLastModifiedTime << Log::end;
     }
     else
     {
         // Check if document has been modified by some external action
-        LOG_TRC("Document modified time: " <<
-                Poco::DateTimeFormatter::format(Poco::DateTime(fileInfo._modifiedTime),
-                                                Poco::DateTimeFormat::ISO8601_FORMAT));
+        Log::trace() << "Document modified time: " << fileInfo._modifiedTime << Log::end;
         static const Poco::Timestamp Zero(Poco::Timestamp::fromEpochTime(0));
         if (_documentLastModifiedTime != Zero &&
             fileInfo._modifiedTime != Zero &&
             _documentLastModifiedTime != fileInfo._modifiedTime)
         {
-            LOG_ERR("Document has been modified behind our back, URI [" << session->getPublicUri().toString() << "].");
-            // What do do?
+            Log::trace() << "Document " << _docKey << "] has been modified behind our back. Informing all clients."
+                         << "Expected: " << _documentLastModifiedTime
+                         << "Actual: " << fileInfo._modifiedTime << Log::end;
+
+            _documentChangedInStorage = true;
+            std::string message = "close: documentconflict";
+            if (_isModified)
+                message = "error: cmd=storage kind=documentconflict";
+
+            session->sendTextFrame(message);
+            broadcastMessage(message);
         }
     }
 
     // Let's load the document now, if not loaded.
     if (!_storage->isLoaded())
     {
-        const auto localPath = _storage->loadStorageFileToLocal(session->getAccessToken());
+        auto localPath = _storage->loadStorageFileToLocal(session->getAuthorization());
+
+        // Check if we have a prefilter "plugin" for this document format
+        for (const auto& plugin : LOOLWSD::PluginConfigurations)
+        {
+            try
+            {
+                std::string extension(plugin->getString("prefilter.extension"));
+                std::string newExtension(plugin->getString("prefilter.newextension"));
+                std::string commandLine(plugin->getString("prefilter.commandline"));
+
+                if (localPath.length() > extension.length()+1 &&
+                    strcasecmp(localPath.substr(localPath.length() - extension.length() -1).data(), (std::string(".") + extension).data()) == 0)
+                {
+                    // Extension matches, try the conversion. We convert the file to another one in
+                    // the same (jail) directory, with just the new extension tacked on.
+
+                    std::string newRootPath = _storage->getRootFilePath() + "." + newExtension;
+
+                    // The commandline must contain the space-separated substring @INPUT@ that is
+                    // replaced with the input file name, and @OUTPUT@ for the output file name.
+
+                    std::vector<std::string>args;
+
+                    Poco::StringTokenizer tokenizer(commandLine, " ");
+                    if (tokenizer.replace("@INPUT@", _storage->getRootFilePath()) != 1 ||
+                        tokenizer.replace("@OUTPUT@", newRootPath) != 1)
+                        throw Poco::NotFoundException();
+
+                    for (std::size_t i = 1; i < tokenizer.count(); i++)
+                        args.push_back(tokenizer[i]);
+
+                    Poco::ProcessHandle process = Poco::Process::launch(tokenizer[0], args);
+                    int rc = process.wait();
+                    if (rc != 0)
+                    {
+                        LOG_ERR("Conversion from " << extension << " to " << newExtension <<" failed");
+                        return false;
+                    }
+                    _storage->setRootFilePath(newRootPath);
+                    localPath += "." + newExtension;
+                }
+
+                // We successfully converted the file to something LO can use; break out of the for
+                // loop.
+                break;
+            }
+            catch (const Poco::NotFoundException&)
+            {
+                // This plugin is not a proper prefilter one
+            }
+        }
 
         std::ifstream istr(localPath, std::ios::binary);
         Poco::SHA1Engine sha1;
@@ -524,6 +627,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         // Use the local temp file's timestamp.
         _lastFileModifiedTime = Poco::File(_storage->getRootFilePath()).getLastModified();
         _tileCache.reset(new TileCache(_storage->getUri(), _lastFileModifiedTime, _cacheRoot));
+        _tileCache->setThreadOwner(std::this_thread::get_id());
     }
 
     LOOLWSD::dumpNewSessionTrace(getJailId(), sessionId, _uriOrig, _storage->getRootFilePath());
@@ -544,10 +648,15 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
 }
 
 bool DocumentBroker::saveToStorage(const std::string& sessionId,
-                                   bool success, const std::string& result)
+                                   bool success, const std::string& result, bool force)
 {
     assertCorrectThread();
 
+    if (force)
+    {
+        LOG_TRC("Document will be saved forcefully to storage.");
+        _storage->forceSave();
+    }
     const bool res = saveToStorageInternal(sessionId, success, result);
 
     // If marked to destroy, or session is disconnected, remove.
@@ -565,11 +674,21 @@ bool DocumentBroker::saveToStorage(const std::string& sessionId,
     return res;
 }
 
-bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
-                                           bool success, const std::string& result)
+bool DocumentBroker::saveAsToStorage(const std::string& sessionId, const std::string& saveAsPath, const std::string& saveAsFilename)
 {
     assertCorrectThread();
 
+    return saveToStorageInternal(sessionId, true, "", saveAsPath, saveAsFilename);
+}
+
+bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
+                                           bool success, const std::string& result,
+                                           const std::string& saveAsPath, const std::string& saveAsFilename)
+{
+    assertCorrectThread();
+
+    const bool isSaveAs = !saveAsPath.empty();
+
     // If save requested, but core didn't save because document was unmodified
     // notify the waiting thread, if any.
     LOG_TRC("Saving to storage docKey [" << _docKey << "] for session [" << sessionId <<
@@ -589,13 +708,21 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
         return false;
     }
 
-    const std::string accessToken = it->second->getAccessToken();
-    const auto uri = it->second->getPublicUri().toString();
+    // Check that we are actually about to upload a successfully saved document.
+    if (!success)
+    {
+        LOG_ERR("Cannot save docKey [" << _docKey << "], the .uno:Save has failed in LOK.");
+        it->second->sendTextFrame("error: cmd=storage kind=savefailed");
+        return false;
+    }
+
+    const Authorization auth = it->second->getAuthorization();
+    const auto uri = isSaveAs? saveAsPath: it->second->getPublicUri().toString();
 
     // If we aren't destroying the last editable session just yet,
     // and the file timestamp hasn't changed, skip saving.
     const auto newFileModifiedTime = Poco::File(_storage->getRootFilePath()).getLastModified();
-    if (!_lastEditableSession && newFileModifiedTime == _lastFileModifiedTime)
+    if (!isSaveAs && !_lastEditableSession && newFileModifiedTime == _lastFileModifiedTime)
     {
         // Nothing to do.
         LOG_DBG("Skipping unnecessary saving to URI [" << uri << "] with docKey [" << _docKey <<
@@ -607,42 +734,45 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
 
     LOG_DBG("Persisting [" << _docKey << "] after saving to URI [" << uri << "].");
 
-    // FIXME: We should check before persisting the document that it hasn't been updated in its
-    // storage behind our backs.
-
     assert(_storage && _tileCache);
-    StorageBase::SaveResult storageSaveResult = _storage->saveLocalFileToStorage(accessToken);
-    if (storageSaveResult == StorageBase::SaveResult::OK)
+    StorageBase::SaveResult storageSaveResult = _storage->saveLocalFileToStorage(auth, saveAsPath, saveAsFilename);
+    if (storageSaveResult.getResult() == StorageBase::SaveResult::OK)
     {
-        _isModified = false;
-        _tileCache->setUnsavedChanges(false);
-        _lastFileModifiedTime = newFileModifiedTime;
-        _tileCache->saveLastModified(_lastFileModifiedTime);
-        _lastSaveTime = std::chrono::steady_clock::now();
-        _poll->wakeup();
-
-        // Calling getWOPIFileInfo() or getLocalFileInfo() has the side-effect of updating
-        // StorageBase::_fileInfo. Get the timestamp of the document as persisted in its storage
-        // from there.
-        // FIXME: Yes, of course we should turn this stuff into a virtual function and avoid this
-        // dynamic_cast dance.
-        if (dynamic_cast<WopiStorage*>(_storage.get()) != nullptr)
+        if (!isSaveAs)
         {
-            auto wopiFileInfo = static_cast<WopiStorage*>(_storage.get())->getWOPIFileInfo(accessToken);
-        }
-        else if (dynamic_cast<LocalStorage*>(_storage.get()) != nullptr)
-        {
-            auto localFileInfo = static_cast<LocalStorage*>(_storage.get())->getLocalFileInfo();
-        }
-        // So set _documentLastModifiedTime then
-        _documentLastModifiedTime = _storage->getFileInfo()._modifiedTime;
+            setModified(false);
+            _lastFileModifiedTime = newFileModifiedTime;
+            _tileCache->saveLastModified(_lastFileModifiedTime);
+            _lastSaveTime = std::chrono::steady_clock::now();
+            _poll->wakeup();
 
-        LOG_DBG("Saved docKey [" << _docKey << "] to URI [" << uri << "] and updated tile cache. Document modified timestamp: " <<
-                Poco::DateTimeFormatter::format(Poco::DateTime(_documentLastModifiedTime),
-                                                               Poco::DateTimeFormat::ISO8601_FORMAT));
+            // So set _documentLastModifiedTime then
+            _documentLastModifiedTime = _storage->getFileInfo()._modifiedTime;
+
+            // After a successful save, we are sure that document in the storage is same as ours
+            _documentChangedInStorage = false;
+
+            Log::debug() << "Saved docKey [" << _docKey << "] to URI [" << uri
+                         << "] and updated tile cache. Document modified timestamp: "
+                         << _documentLastModifiedTime << Log::end;
+        }
+        else
+        {
+            // normalize the url (mainly to " " -> "%20")
+            std::string url = Poco::URI(storageSaveResult.getSaveAsUrl()).toString();
+
+            // encode the name
+            std::string encodedName;
+            Poco::URI::encode(storageSaveResult.getSaveAsName(), "", encodedName);
+
+            it->second->sendTextFrame("saveas: url=" + url + " filename=" + encodedName);
+
+            Log::debug() << "Saved As docKey [" << _docKey << "] to URI [" << url
+                         << " with name '" << encodedName << "'] successfully.";
+        }
         return true;
     }
-    else if (storageSaveResult == StorageBase::SaveResult::DISKFULL)
+    else if (storageSaveResult.getResult() == StorageBase::SaveResult::DISKFULL)
     {
         LOG_WRN("Disk full while saving docKey [" << _docKey << "] to URI [" << uri <<
                 "]. Making all sessions on doc read-only and notifying clients.");
@@ -654,17 +784,30 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId,
             sessionIt.second->sendTextFrame("error: cmd=storage kind=savediskfull");
         }
     }
-    else if (storageSaveResult == StorageBase::SaveResult::UNAUTHORIZED)
+    else if (storageSaveResult.getResult() == StorageBase::SaveResult::UNAUTHORIZED)
     {
         LOG_ERR("Cannot save docKey [" << _docKey << "] to storage URI [" << uri << "]. Invalid or expired access token. Notifying client.");
         it->second->sendTextFrame("error: cmd=storage kind=saveunauthorized");
     }
-    else if (storageSaveResult == StorageBase::SaveResult::FAILED)
+    else if (storageSaveResult.getResult() == StorageBase::SaveResult::FAILED)
     {
         //TODO: Should we notify all clients?
         LOG_ERR("Failed to save docKey [" << _docKey << "] to URI [" << uri << "]. Notifying client.");
         it->second->sendTextFrame("error: cmd=storage kind=savefailed");
     }
+    else if (storageSaveResult.getResult() == StorageBase::SaveResult::DOC_CHANGED)
+    {
+        LOG_ERR("PutFile says that Document changed in storage");
+        _documentChangedInStorage = true;
+        if (_isModified)
+        {
+            broadcastMessage("error: cmd=storage kind=documentconflict");
+        }
+        else
+        {
+            closeDocument("documentconflict");
+        }
+    }
 
     return false;
 }
@@ -727,19 +870,21 @@ bool DocumentBroker::autoSave(const bool force)
         LOG_TRC("Time since last save of docKey [" << _docKey << "] is " << timeSinceLastSaveMs <<
                 "ms and most recent activity was " << inactivityTimeMs << "ms ago.");
 
+        static const auto idleSaveDurationMs = LOOLWSD::getConfigValue<int>("per_document.idlesave_duration_secs", 30) * 1000;
+        static const auto autoSaveDurationMs = LOOLWSD::getConfigValue<int>("per_document.autosave_duration_secs", 300) * 1000;
         // Either we've been idle long enough, or it's auto-save time.
-        if (inactivityTimeMs >= IdleSaveDurationMs ||
-            timeSinceLastSaveMs >= AutoSaveDurationMs)
+        if (inactivityTimeMs >= idleSaveDurationMs ||
+            timeSinceLastSaveMs >= autoSaveDurationMs)
         {
             LOG_TRC("Sending timed save command for [" << _docKey << "].");
-            sent = sendUnoSave(savingSessionId);
+            sent = sendUnoSave(savingSessionId, /*dontTerminateEdit=*/ true, /*dontSaveIfUnmodified=*/ true, /*isAutosave=*/ true);
         }
     }
 
     return sent;
 }
 
-bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminateEdit, bool dontSaveIfUnmodified)
+bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminateEdit, bool dontSaveIfUnmodified, bool isAutosave)
 {
     assertCorrectThread();
 
@@ -780,6 +925,10 @@ bool DocumentBroker::sendUnoSave(const std::string& sessionId, bool dontTerminat
         // arguments end
         oss << "}";
 
+        assert(_storage);
+        _storage->setUserModified(_isModified);
+        _storage->setIsAutosave(isAutosave || UnitWSD::get().isAutosave());
+
         const auto saveArgs = oss.str();
         LOG_TRC(".uno:Save arguments: " << saveArgs);
         const auto command = "uno .uno:Save " + saveArgs;
@@ -856,7 +1005,7 @@ size_t DocumentBroker::addSessionInternal(const std::shared_ptr<ClientSession>&
     _childProcess->sendTextFrame(aMessage);
 
     // Tell the admin console about this new doc
-    Admin::instance().addDoc(_docKey, getPid(), getFilename(), id, session->getUserName());
+    Admin::instance().addDoc(_docKey, getPid(), getFilename(), id, session->getUserName(), session->getUserId());
 
     // Add and attach the session.
     _sessions.emplace(session->getId(), session);
@@ -955,6 +1104,9 @@ void DocumentBroker::alertAllUsers(const std::string& msg)
 {
     assertCorrectThread();
 
+    if (UnitWSD::get().filterAlertAllusers(msg))
+        return;
+
     auto payload = std::make_shared<Message>(msg, Message::Dir::Out);
 
     LOG_DBG("Alerting all users of [" << _docKey << "]: " << msg);
@@ -1251,14 +1403,25 @@ void DocumentBroker::destroyIfLastEditor(const std::string& id)
 
 void DocumentBroker::setModified(const bool value)
 {
+    if(_isModified != value)
+    {
+        _isModified = value;
+        Admin::instance().modificationAlert(_docKey, getPid(), value);
+    }
+
     _tileCache->setUnsavedChanges(value);
-    _isModified = value;
 }
 
 bool DocumentBroker::forwardToChild(const std::string& viewId, const std::string& message)
 {
     assertCorrectThread();
 
+    // Ignore userinactive, useractive message until document is loaded
+    if (!isLoaded() && (message == "userinactive" || message == "useractive"))
+    {
+        return true;
+    }
+
     LOG_TRC("Forwarding payload to child [" << viewId << "]: " << message);
 
     std::string msg = "child-" + viewId + ' ' + message;
@@ -1339,7 +1502,7 @@ bool DocumentBroker::forwardToClient(const std::shared_ptr<Message>& payload)
 void DocumentBroker::shutdownClients(const std::string& closeReason)
 {
     assertCorrectThread();
-    LOG_INF("Terminating " << _sessions.size() << " clients of doc [" << _docKey << "].");
+    LOG_INF("Terminating " << _sessions.size() << " clients of doc [" << _docKey << "] with reason: " << closeReason);
 
     // First copy into local container, since removeSession
     // will erase from _sessions, but will leave the last.
@@ -1377,17 +1540,14 @@ void DocumentBroker::childSocketTerminated()
     shutdownClients("terminated");
 }
 
-void DocumentBroker::terminateChild(const std::string& closeReason, const bool rude)
+void DocumentBroker::terminateChild(const std::string& closeReason)
 {
     assertCorrectThread();
 
-    LOG_INF("Terminating doc [" << _docKey << "].");
+    LOG_INF("Terminating doc [" << _docKey << "] with reason: " << closeReason);
 
     // Close all running sessions
-    if (!rude)
-    {
-        shutdownClients(closeReason);
-    }
+    shutdownClients(closeReason);
 
     if (_childProcess)
     {
@@ -1395,12 +1555,8 @@ void DocumentBroker::terminateChild(const std::string& closeReason, const bool r
 
         // First flag to stop as it might be waiting on our lock
         // to process some incoming message.
-        if (!rude)
-        {
-            _childProcess->stop();
-        }
-
-        _childProcess->close(rude);
+        _childProcess->stop();
+        _childProcess->close(false);
     }
 
     _stop = true;
@@ -1411,7 +1567,19 @@ void DocumentBroker::closeDocument(const std::string& reason)
     assertCorrectThread();
 
     LOG_DBG("Closing DocumentBroker for docKey [" << _docKey << "] with reason: " << reason);
-    terminateChild(reason, true);
+    _closeReason = reason; // used later in the polling loop
+    stop();
+}
+
+void DocumentBroker::broadcastMessage(const std::string& message)
+{
+    assertCorrectThread();
+
+    LOG_DBG("Broadcasting message [" << message << "] to all sessions.");
+    for (const auto& sessionIt : _sessions)
+    {
+        sessionIt.second->sendTextFrame(message);
+    }
 }
 
 void DocumentBroker::updateLastActivityTime()
@@ -1420,10 +1588,27 @@ void DocumentBroker::updateLastActivityTime()
     Admin::instance().updateLastActivityTime(_docKey);
 }
 
+void DocumentBroker::getIOStats(uint64_t &sent, uint64_t &recv)
+{
+    sent = 0;
+    recv = 0;
+    assertCorrectThread();
+    for (const auto& sessionIt : _sessions)
+    {
+        uint64_t s, r;
+        sessionIt.second->getIOStats(s, r);
+        sent += s;
+        recv += r;
+    }
+}
+
 void DocumentBroker::dumpState(std::ostream& os)
 {
     std::unique_lock<std::mutex> lock(_mutex);
 
+    uint64_t sent, recv;
+    getIOStats(sent, recv);
+
     os << " Broker: " << _filename << " pid: " << getPid();
     if (_markToDestroy)
         os << " *** Marked to destroy ***";
@@ -1433,6 +1618,8 @@ void DocumentBroker::dumpState(std::ostream& os)
         os << "\n  loaded in: " << _loadDuration.count() << "ms";
     else
         os << "\n  still loading...";
+    os << "\n  sent: " << sent;
+    os << "\n  recv: " << recv;
     os << "\n  modified?: " << _isModified;
     os << "\n  jail id: " << _jailId;
     os << "\n  filename: " << _filename;
