@@ -66,14 +66,6 @@ public:
     using LoolException::LoolException;
 };
 
-/// An access-denied exception that is meant to signify
-/// a storage authentication error.
-class AccessDeniedException : public LoolException
-{
-public:
-    using LoolException::LoolException;
-};
-
 /// A service-unavailable exception that is meant to signify
 /// an internal error.
 class ServiceUnavailableException : public LoolException
