@@ -7,16 +7,16 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "TileCache.hpp"
 #include "config.h"
 
+#include "TileCache.hpp"
+
 #include <cassert>
 #include <climits>
 #include <cstdio>
 #include <fstream>
 #include <iostream>
 #include <memory>
-#include <mutex>
 #include <sstream>
 #include <string>
 #include <vector>
@@ -32,29 +32,29 @@
 
 #include "ClientSession.hpp"
 #include "Common.hpp"
-#include "common/FileUtil.hpp"
 #include "Protocol.hpp"
+#include "SenderQueue.hpp"
 #include "Unit.hpp"
 #include "Util.hpp"
+#include "common/FileUtil.hpp"
+
+using namespace LOOLProtocol;
 
 using Poco::DirectoryIterator;
 using Poco::File;
 using Poco::StringTokenizer;
 using Poco::Timestamp;
 
-using namespace LOOLProtocol;
-
 TileCache::TileCache(const std::string& docURL,
                      const Timestamp& modifiedTime,
                      const std::string& cacheDir) :
     _docURL(docURL),
-    _cacheDir(cacheDir),
-    _cacheMutex(),
-    _tilesBeingRenderedMutex()
+    _cacheDir(cacheDir)
 {
-    Log::info() << "TileCache ctor for uri [" << _docURL
-                << "] modifiedTime=" << (modifiedTime.raw()/1000000)
-                << " getLastModified()=" << (getLastModified().raw()/1000000) << Log::end;
+    LOG_INF("TileCache ctor for uri [" << _docURL <<
+            "], cacheDir: [" << _cacheDir <<
+            "], modifiedTime=" << (modifiedTime.raw()/1000000) <<
+            " getLastModified()=" << (getLastModified().raw()/1000000));
     File directory(_cacheDir);
     std::string unsaved;
     if (directory.exists() &&
@@ -62,8 +62,7 @@ TileCache::TileCache(const std::string& docURL,
          getTextFile("unsaved.txt", unsaved)))
     {
         // Document changed externally or modifications were not saved after all. Cache not useful.
-        FileUtil::removeFile(_cacheDir, true);
-        Log::info("Completely cleared tile cache: " + _cacheDir);
+        completeCleanup();
     }
 
     File(_cacheDir).createDirectories();
@@ -73,7 +72,14 @@ TileCache::TileCache(const std::string& docURL,
 
 TileCache::~TileCache()
 {
-    Log::info("~TileCache dtor for uri [" + _docURL + "].");
+    _owner = std::thread::id(0);
+    LOG_INF("~TileCache dtor for uri [" << _docURL << "].");
+}
+
+void TileCache::completeCleanup() const
+{
+    FileUtil::removeFile(_cacheDir, true);
+    LOG_INF("Completely cleared tile cache: " << _cacheDir);
 }
 
 /// Tracks the rendering of a given tile
@@ -98,11 +104,6 @@ struct TileCache::TileBeingRendered
     std::chrono::steady_clock::time_point getStartTime() const { return _startTime; }
     double getElapsedTimeMs() const { return std::chrono::duration_cast<std::chrono::milliseconds>
                                               (std::chrono::steady_clock::now() - _startTime).count(); }
-    void resetStartTime()
-    {
-        _startTime = std::chrono::steady_clock::now();
-    }
-
 private:
     std::chrono::steady_clock::time_point _startTime;
     TileDesc _tile;
@@ -113,17 +114,17 @@ std::shared_ptr<TileCache::TileBeingRendered> TileCache::findTileBeingRendered(c
 {
     const std::string cachedName = cacheFileName(tileDesc);
 
-    Util::assertIsLocked(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     const auto tile = _tilesBeingRendered.find(cachedName);
-    return (tile != _tilesBeingRendered.end() ? tile->second : nullptr);
+    return tile != _tilesBeingRendered.end() ? tile->second : nullptr;
 }
 
 void TileCache::forgetTileBeingRendered(const TileDesc& tile)
 {
     const std::string cachedName = cacheFileName(tile);
 
-    Util::assertIsLocked(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     _tilesBeingRendered.erase(cachedName);
 }
@@ -139,7 +140,7 @@ std::unique_ptr<std::fstream> TileCache::lookupTile(const TileDesc& tile)
 
     if (result && result->is_open())
     {
-        Log::trace("Found cache tile: " + fileName);
+        LOG_TRC("Found cache tile: " << fileName);
         return result;
     }
 
@@ -148,7 +149,7 @@ std::unique_ptr<std::fstream> TileCache::lookupTile(const TileDesc& tile)
 
 void TileCache::saveTileAndNotify(const TileDesc& tile, const char *data, const size_t size)
 {
-    std::unique_lock<std::mutex> lock(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     std::shared_ptr<TileBeingRendered> tileBeingRendered = findTileBeingRendered(tile);
 
@@ -156,81 +157,74 @@ void TileCache::saveTileAndNotify(const TileDesc& tile, const char *data, const
     const auto cachedName = (tileBeingRendered ? tileBeingRendered->getCacheName()
                                                : cacheFileName(tile));
 
-    // Ignore if we can't save the tile, things will work anyway, but slower. An error indication
-    // has been supposed to be sent to all users in that case.
+    // Ignore if we can't save the tile, things will work anyway, but slower.
+    // An error indication is supposed to be sent to all users in that case.
     const auto fileName = _cacheDir + "/" + cachedName;
     if (FileUtil::saveDataToFileSafely(fileName, data, size))
     {
-        Log::trace() << "Saved cache tile: " << fileName << Log::end;
+        LOG_TRC("Saved cache tile: " << fileName);
     }
 
     // Notify subscribers, if any.
     if (tileBeingRendered)
     {
-        if (!tileBeingRendered->_subscribers.empty())
+        const auto subscriberCount = tileBeingRendered->_subscribers.size();
+        if (subscriberCount > 0)
         {
             std::string response = tile.serialize("tile:");
-            Log::debug("Sending tile message to subscribers: " + response);
-
-            std::vector<char> output(256 + size);
-            output.resize(response.size() + 1 + size);
-
-            std::memcpy(output.data(), response.data(), response.size());
-            output[response.size()] = '\n';
-            std::memcpy(output.data() + response.size() + 1, data, size);
+            LOG_DBG("Sending tile message to " << subscriberCount << " subscribers: " << response);
 
             // Send to first subscriber as-is (without cache marker).
-            auto firstSubscriber = tileBeingRendered->_subscribers[0].lock();
-            if (firstSubscriber)
-            {
-                try
-                {
-                    firstSubscriber->sendBinaryFrame(output.data(), output.size());
-                }
-                catch (const std::exception& ex)
-                {
-                    Log::warn("Failed to send tile to " + firstSubscriber->getName() + ": " + ex.what());
-                }
-            }
+            auto payload = std::make_shared<Message>(response,
+                                                     Message::Dir::Out,
+                                                     response.size() + 1 + size);
+            payload->append("\n", 1);
+            payload->append(data, size);
 
-            // All others must get served from the cache.
-            response += " renderid=cached\n";
-            output.resize(response.size() + size);
-            std::memcpy(output.data(), response.data(), response.size());
-            std::memcpy(output.data() + response.size(), data, size);
+            auto& firstSubscriber = tileBeingRendered->_subscribers[0];
+            auto firstSession = firstSubscriber.lock();
+            if (firstSession)
+                firstSession->enqueueSendMessage(payload);
 
-            for (size_t i = 1; i < tileBeingRendered->_subscribers.size(); ++i)
+            if (subscriberCount > 1)
             {
-                auto subscriber = tileBeingRendered->_subscribers[i].lock();
-                if (subscriber)
+                // All others must get served from the cache.
+                response += " renderid=cached\n";
+
+                // Create a new Payload.
+                payload.reset();
+                payload = std::make_shared<Message>(response,
+                                                    Message::Dir::Out,
+                                                    response.size() + size);
+                payload->append(data, size);
+
+                for (size_t i = 1; i < subscriberCount; ++i)
                 {
-                    try
+                    auto& subscriber = tileBeingRendered->_subscribers[i];
+                    auto session = subscriber.lock();
+                    if (session)
                     {
-                        subscriber->sendBinaryFrame(output.data(), output.size());
-                    }
-                    catch (const std::exception& ex)
-                    {
-                        Log::warn("Failed to send tile to " + subscriber->getName() + ": " + ex.what());
+                        session->enqueueSendMessage(payload);
                     }
                 }
             }
         }
         else
         {
-            Log::debug("No subscribers for: " + cachedName);
+            LOG_DBG("No subscribers for: " << cachedName);
         }
 
         // Remove subscriptions.
         if (tileBeingRendered->getVersion() <= tile.getVersion())
         {
-            Log::debug() << "STATISTICS: tile " << tile.getVersion() << " internal roundtrip "
-                         << tileBeingRendered->getElapsedTimeMs() << " ms." << Log::end;
+            LOG_DBG("STATISTICS: tile " << tile.getVersion() << " internal roundtrip " <<
+                    tileBeingRendered->getElapsedTimeMs() << " ms.");
             _tilesBeingRendered.erase(cachedName);
         }
     }
     else
     {
-        Log::debug("No subscribers for: " + cachedName);
+        LOG_DBG("No subscribers for: " << cachedName);
     }
 }
 
@@ -241,7 +235,7 @@ bool TileCache::getTextFile(const std::string& fileName, std::string& content)
     std::fstream textStream(fullFileName, std::ios::in);
     if (!textStream.is_open())
     {
-        Log::info("Could not open " + fullFileName);
+        LOG_INF("Could not open " << fullFileName);
         return false;
     }
 
@@ -257,7 +251,8 @@ bool TileCache::getTextFile(const std::string& fileName, std::string& content)
         buffer.pop_back();
 
     content = std::string(buffer.data(), buffer.size());
-    Log::info("Read '" + LOOLProtocol::getAbbreviatedMessage(content.c_str(), content.size()) + "' from " + fullFileName);
+    LOG_INF("Read '" << LOOLProtocol::getAbbreviatedMessage(content.c_str(), content.size()) <<
+            "' from " << fullFileName);
 
     return true;
 }
@@ -269,12 +264,13 @@ void TileCache::saveTextFile(const std::string& text, const std::string& fileNam
 
     if (!textStream.is_open())
     {
-        Log::error("Could not save '" + text + "' to " + fullFileName);
+        LOG_ERR("Could not save '" << text << "' to " << fullFileName);
         return;
     }
     else
     {
-        Log::info("Saving '" + LOOLProtocol::getAbbreviatedMessage(text.c_str(), text.size()) + "' to " + fullFileName);
+        LOG_INF("Saving '" << LOOLProtocol::getAbbreviatedMessage(text.c_str(), text.size()) <<
+                "' to " << fullFileName);
     }
 
     textStream << text << std::endl;
@@ -318,15 +314,14 @@ std::unique_ptr<std::fstream> TileCache::lookupCachedFile(const std::string& nam
 
 void TileCache::invalidateTiles(int part, int x, int y, int width, int height)
 {
-    Log::trace() << "Removing invalidated tiles: part: " << part
-                 << ", x: " << x << ", y: " << y
-                 << ", width: " << width
-                 << ", height: " << height << Log::end;
+    LOG_TRC("Removing invalidated tiles: part: " << part <<
+            ", x: " << x << ", y: " << y <<
+            ", width: " << width <<
+            ", height: " << height);
 
     File dir(_cacheDir);
 
-    std::unique_lock<std::mutex> lock(_cacheMutex);
-    std::unique_lock<std::mutex> lockSubscribers(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     if (dir.exists() && dir.isDirectory())
     {
@@ -335,7 +330,7 @@ void TileCache::invalidateTiles(int part, int x, int y, int width, int height)
             const std::string fileName = tileIterator.path().getFileName();
             if (intersectsTile(fileName, part, x, y, width, height))
             {
-                Log::debug("Removing tile: " + tileIterator.path().toString());
+                LOG_DBG("Removing tile: " << tileIterator.path().toString());
                 FileUtil::removeFile(tileIterator.path());
             }
         }
@@ -377,7 +372,7 @@ void TileCache::invalidateTiles(const std::string& tiles)
         }
     }
 
-    Log::error("Unexpected invalidatetiles request: " + tiles);
+    LOG_ERR("Unexpected invalidatetiles request [" << tiles << "].");
 }
 
 void TileCache::removeFile(const std::string& fileName)
@@ -385,7 +380,7 @@ void TileCache::removeFile(const std::string& fileName)
     const std::string fullFileName = _cacheDir + "/" + fileName;
 
     if (std::remove(fullFileName.c_str()) == 0)
-        Log::info("Removed file: " + fullFileName);
+        LOG_INF("Removed file: " << fullFileName);
 }
 
 std::string TileCache::cacheFileName(const TileDesc& tile)
@@ -397,12 +392,12 @@ std::string TileCache::cacheFileName(const TileDesc& tile)
     return oss.str();
 }
 
-bool TileCache::parseCacheFileName(const std::string& fileName, int& part, int& width, int& height, int& tilePosX, int& tilePosY, int& tileWidth, int& tileHeight) const
+bool TileCache::parseCacheFileName(const std::string& fileName, int& part, int& width, int& height, int& tilePosX, int& tilePosY, int& tileWidth, int& tileHeight)
 {
-    return (std::sscanf(fileName.c_str(), "%d_%dx%d.%d,%d.%dx%d.png", &part, &width, &height, &tilePosX, &tilePosY, &tileWidth, &tileHeight) == 7);
+    return std::sscanf(fileName.c_str(), "%d_%dx%d.%d,%d.%dx%d.png", &part, &width, &height, &tilePosX, &tilePosY, &tileWidth, &tileHeight) == 7;
 }
 
-bool TileCache::intersectsTile(const std::string& fileName, int part, int x, int y, int width, int height) const
+bool TileCache::intersectsTile(const std::string& fileName, int part, int x, int y, int width, int height)
 {
     int tilePart, tilePixelWidth, tilePixelHeight, tilePosX, tilePosY, tileWidth, tileHeight;
     if (parseCacheFileName(fileName, tilePart, tilePixelWidth, tilePixelHeight, tilePosX, tilePosY, tileWidth, tileHeight))
@@ -444,15 +439,13 @@ void TileCache::saveLastModified(const Timestamp& timestamp)
 }
 
 // FIXME: to be further simplified when we centralize tile messages.
-void TileCache::subscribeToTileRendering(const TileDesc& tile, const std::shared_ptr<ClientSession> &subscriber)
+void TileCache::subscribeToTileRendering(const TileDesc& tile, const std::shared_ptr<ClientSession>& subscriber)
 {
-    assert(subscriber->getKind() == LOOLSession::Kind::ToClient);
-
     std::ostringstream oss;
     oss << '(' << tile.getPart() << ',' << tile.getTilePosX() << ',' << tile.getTilePosY() << ')';
     const auto name = oss.str();
 
-    std::unique_lock<std::mutex> lock(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     std::shared_ptr<TileBeingRendered> tileBeingRendered = findTileBeingRendered(tile);
 
@@ -462,15 +455,14 @@ void TileCache::subscribeToTileRendering(const TileDesc& tile, const std::shared
         {
             if (s.lock().get() == subscriber.get())
             {
-                Log::debug("Redundant request to subscribe on tile " + name);
+                LOG_DBG("Redundant request to subscribe on tile " << name);
                 tileBeingRendered->setVersion(tile.getVersion());
                 return;
             }
         }
 
-        Log::debug() << "Subscribing to tile " << name << " which has "
-                     << tileBeingRendered->_subscribers.size()
-                     << " subscribers already. Adding one more." << Log::end;
+        LOG_DBG("Subscribing " << subscriber->getName() << " to tile " << name << " which has " <<
+                tileBeingRendered->_subscribers.size() << " subscribers already.");
         tileBeingRendered->_subscribers.push_back(subscriber);
 
         const auto duration = (std::chrono::steady_clock::now() - tileBeingRendered->getStartTime());
@@ -482,8 +474,8 @@ void TileCache::subscribeToTileRendering(const TileDesc& tile, const std::shared
     }
     else
     {
-        Log::debug() << "Subscribing to tile " << name << " which has no subscribers. Subscribing for ver: "
-                     << tile.getVersion() << "." << Log::end;
+        LOG_DBG("Subscribing " << subscriber->getName() << " to tile " << name <<
+                " ver=" << tile.getVersion() << " which has no subscribers.");
 
         const std::string cachedName = cacheFileName(tile);
 
@@ -498,9 +490,9 @@ void TileCache::subscribeToTileRendering(const TileDesc& tile, const std::shared
 std::string TileCache::cancelTiles(const std::shared_ptr<ClientSession> &subscriber)
 {
     assert(subscriber && "cancelTiles expects valid subscriber");
-    Log::trace("Cancelling tiles for " + subscriber->getName());
+    LOG_TRC("Cancelling tiles for " << subscriber->getName());
 
-    std::unique_lock<std::mutex> lock(_tilesBeingRenderedMutex);
+    assertCorrectThread();
 
     const auto sub = subscriber.get();
 
@@ -516,13 +508,14 @@ std::string TileCache::cancelTiles(const std::shared_ptr<ClientSession> &subscri
         }
 
         auto& subscribers = it->second->_subscribers;
-        Log::trace("Tile " + it->first + " has " + std::to_string(subscribers.size()) + " subscribers.");
+        LOG_TRC("Tile " << it->first << " has " << subscribers.size() << " subscribers.");
 
         const auto itRem = std::find_if(subscribers.begin(), subscribers.end(),
                                         [sub](std::weak_ptr<ClientSession>& ptr){ return ptr.lock().get() == sub; });
         if (itRem != subscribers.end())
         {
-            Log::trace("Tile " + it->first + " has " + std::to_string(subscribers.size()) + " subscribers. Removing one.");
+            LOG_TRC("Tile " << it->first << " has " << subscribers.size() <<
+                    " subscribers. Removing " << subscriber->getName() << ".");
             subscribers.erase(itRem, itRem + 1);
             if (subscribers.empty())
             {
@@ -537,7 +530,17 @@ std::string TileCache::cancelTiles(const std::shared_ptr<ClientSession> &subscri
     }
 
     const auto canceltiles = oss.str();
-    return (canceltiles.empty() ? canceltiles : "canceltiles " + canceltiles);
+    return canceltiles.empty() ? canceltiles : "canceltiles " + canceltiles;
+}
+
+void TileCache::assertCorrectThread()
+{
+    const bool correctThread = _owner == std::thread::id(0) || std::this_thread::get_id() == _owner;
+    if (!correctThread)
+        LOG_ERR("TileCache method invoked from foreign thread. Expected: 0x" << std::hex <<
+                _owner << " but called from 0x" << std::this_thread::get_id() << " (" <<
+                std::dec << Util::getThreadId() << ").");
+    assert (correctThread);
 }
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
